package guru.kake.ronove

import javafx.beans.property.SimpleBooleanProperty
import javafx.stage.FileChooser
import tornadofx.FileChooserMode
import tornadofx.chooseFile
import java.io.File
import java.io.PrintWriter
import kotlin.reflect.KProperty1

/**
 * Class for holding the data of the entire translation project, which consists of pages consisting of lines.
 */
class Project {
    /** Last used file for saving, updated on Load/Save as */
    private var projectFile: File? = null

    /** Observable denoting if projectfile is set */
    var projectFileSet = SimpleBooleanProperty(false)

    /** Pages of the project */
    var pages: MutableList<Page> = mutableListOf(Page())

    /** Prompts user for project file to open and loads it */
    fun open() {
        val file = chooseFile("Open...",
            arrayOf(
                FileChooser.ExtensionFilter("Ronove project files", "*.rvp", "*.RVP"),
                FileChooser.ExtensionFilter("All files", "*")
            ),
            mode=FileChooserMode.Single
        )
        if (file.isEmpty()) return
        loadFile(file[0])
        projectFile = file[0]
        projectFileSet.value = true
    }

    /** Prompts user for project file to open and saves to it */
    fun saveAs() {
        val file = chooseFile("Save as..",
            arrayOf(
                FileChooser.ExtensionFilter("Ronove project files", "*.rvp", "*.RVP"),
                FileChooser.ExtensionFilter("All files", "*")
            ),
            mode=FileChooserMode.Save
        )
        if (file.isEmpty()) return
        saveFile(file[0])
        projectFile = file[0]
        projectFileSet.value = true
    }

    fun save() = projectFile?.let { saveFile(it) }

    fun new() {
        pages = mutableListOf()
        projectFile = null
        projectFileSet.value = false
    }

    private fun loadFile(file: File) {
        if (!file.exists()) return
        new()
        file.bufferedReader().use { reader ->
            var line: String?
            while (reader.readLine().also { line = it } != null) {
                println(line)
                if (line!!.startsWith('#')) {
                    pages.add(Page())
                } else {
                    pages.last().lines.add(Line(line!!))
                }
            }
        }
    }

    private fun saveFile(file: File) {
        file.printWriter().use { writer ->
            var pageNumber = 1
            for (page in pages) {
                writer.println("#${pageNumber++}")
                page.save(writer)
            }
        }
    }
}

/**
 * Class for holding the data of a single translation page
 */
class Page(lines: Collection<Line> = listOf()) {
    /** Lines on a page */
    var lines: MutableList<Line> = lines.toMutableList()

    /**
     * Get a specific column of the page as a list
     * @param column the column to get, for example [Line::original]
     * @return list containing the elements of the column
     */
    fun getColumn(column: KProperty1<Line, String>) : List<String> = buildList {
        for (line in lines) add(column.get(line))
    }

    /**
     * Translates all lines on page
     */
    fun translatePage() {
        TranslationBackend(getColumn(Line::original)).zip(lines).forEach { (translation, line) ->
            line.machineTranslation = translation
        }
    }

    /**
     * Debug function for printing a page in stdout
     */
    @Deprecated("For debug use only", level = DeprecationLevel.WARNING)
    fun debugPrint() {
        for (line in lines) {
            println("${line.original}: ${line.machineTranslation}")
        }
    }

    fun save(writer: PrintWriter) {
        for (line in lines) {
            with(line) {
                writer.println("$original|$machineTranslation|$finalTranslation|$notes")
            }
        }
    }
}

/**
 * Class for holding the data of a single line of translation
 */
data class Line(
    /** Original, untranslated text */
    var original: String = "",
    /** Machine translation of the original */
    var machineTranslation: String = "",
    /** Manual, finalized translation*/
    var finalTranslation: String = "",
    /** Translator notes for the line */
    var notes: String = ""
) {
    constructor(s: String) : this() {
        val args = s.split('|')
        original = args.getOrElse(0) { "" }
        machineTranslation = args.getOrElse(1) { "" }
        finalTranslation = args.getOrElse(2) { "" }
        notes = args.getOrElse(3) { "" }
    }
}